#include "sl45.h"
#include "reg167.h"
#include "math.h"

//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------
#define MAP_W	32
#define SCR_W	12
#define MAP_H	256
#define SCR_H	64

//.    / 8,   
//    
// - 256  
#define MAX_DELTA	32

static const char BSDAT_FILENAME[32]="A:\\bs.dat";
static const char BSLOG_FILENAME[32]="A:\\bs.log";
static const char MAP_FILENAME[32]="A:\\kharkov.map";

typedef unsigned short WORD;

int md_onKey(void far *data, struct MSG far *Msg);
void md_onCreate(void far *data);
void md_onClose(void);

WORD IsFocus(void far *data)
{
 if (((int far *)data)[0x10]) return 0; //  
 if (((int far *)data)[0x0A]) return 0; //   
 if (((int far *)data)[6]!=GetBotDialogWSK_ID()) return 0; //  
 return(1);
}

static const unsigned int md_tab[]={0xFFF5};

static const struct DIALOG_WSK maindialog=
{
 md_onKey,
 md_onCreate,
 md_onClose,
 0x2C,
 sizeof(md_tab)/sizeof(unsigned int),
 md_tab
};

//    
struct BS
{
 unsigned int lac; //LAC
 // CI   10,    
 unsigned int ci; //Cell ID
 unsigned int x; //   x
 unsigned int y; //   y
 char s[32]; //
 unsigned int lvl; // 
};

#define BS_CACHE_SIZE 32

//==========================================================================================
struct VARS
{
 unsigned long tmr;
 unsigned int tmr_online;
 int old_lock_mode;
 //   
 int rollmode;
 unsigned int ready_for_exit; //   *
 //   (  )
 unsigned int viewx;
 unsigned int viewy;
 //   (0x82000)
 unsigned int mapx;
 unsigned int mapy;
 //  
 //-1 -  
 //0 -  
 //1 - 
 //2 -     
 int update_scr;
 struct BS bs[BS_CACHE_SIZE];
 unsigned int bs_cp; //  .     
 int loadCI_in_progress;
 unsigned int oldlac;
 unsigned int oldci;
 unsigned int loadlac;
 unsigned int loadci;
 unsigned int cdraw;
 unsigned int c3draw;
 struct BS far *BS7[7];
 // 
 char celltext[50];
 char aCIupdate[128];
 // ,   
 unsigned int xsize;
 unsigned int ysize;
 char far * mem8;			//buf8k[8192];
 //   
 WSTR up;
 unsigned int us[40];
 char dobj[0x1A];
 //  
 unsigned long sumx;
 unsigned long sumy;
 unsigned int sumk;
 unsigned int gpsx;
 unsigned int gpsy;
 unsigned int manual_x;
 unsigned int manual_y;
 int track_upd;
 int mf_pos; //  
 int mf_state;
 char digits[10]; //  
};
//==========================================================================================
static const unsigned int PWRS[]=
{
 9362, 8838, 8344, 7877, 7436, 7020, 6627, 6257, 5907, 5576,
 5264, 4970, 4692, 4429, 4181, 3948, 3727, 3518, 3321, 3136,
 2960, 2795, 2638, 2491, 2351, 2220, 2095, 1978, 1868, 1763,
 1664, 1571, 1483, 1400, 1322, 1248, 1178, 1112, 1050,  991,
  936,  883,  834,  787,  743,  702,  662,  625,  590,  557,
  526,  497,  469,  442,  418,  394,  372,  351,  332,  313,
  296,  279,  263,  249,  235,  222,  209,  197,  186,  176,
  166,  157,  148,  140,  132,  124,  117,  111,  105,   99,
   93,   88,   83,   78,   74,   70,   66,   62,   59,   55,
   52,   49,   46,   44,   41,   39,   37,   35,   33,   31,
   29,   27,   26,   24,   23,   22,   20,   19,   18,   17,
   16,   15,   14,   14,   13,   12,   11,   11,   10,    9,
    9,    8,    8,    7,    7,    7,    6,    6,    5,    5,
    5,    4,    4,    4,    4,    3,    3,    3,    3,    3,
    2,    2,    2,    2,    2,    2,    2,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    0
/*    ,0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0*/
};

static const char sprites[]=
{
  0xff,0x00,0xff,0x00,
  0xfc,0x00,0x3f,0x00,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0x80,0x01,0x01,0x80,
  0x80,0x3f,0x01,0xfc,
  0x80,0x3f,0x01,0xfc,
  0x80,0x01,0x01,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x01,0x3f,0x80,
  0xfc,0x00,0x3f,0x00,
  0xff,0x00,0xff,0x00,
  0xfc,0x00,0x3f,0x00,
  0xf8,0x01,0x1f,0x80,
  0xf0,0x03,0x0f,0xc0,
  0xf0,0x07,0x0f,0xe0,
  0xc0,0x00,0x03,0x00,
  0x87,0x10,0xe1,0x08,
  0x07,0x30,0xe0,0x0c,
  0x07,0x70,0xe0,0x0e,
  0x07,0x70,0xe0,0x0e,
  0x07,0x30,0xe0,0x0c,
  0x87,0x10,0xe1,0x08,
  0xc0,0x00,0x03,0x00,
  0xf0,0x07,0x0f,0xe0,
  0xf0,0x03,0x0f,0xc0,
  0xf8,0x01,0x1f,0x80,
  0xfc,0x00,0x3f,0x00
};

//    
static const char mf_table[256]=
 "11111111111111111111111111111111"
 "10001**0***0000*012345678900***0"
 "0222333444555666777788899991*110"
 "122233344455566677778889999111*1"
 "11111111111111111111111111111111"
 "11111111111111111111111111111111"
 "22223333444455556666777788889999"
 "22223333444455556666777788889999";

//==========================================================================================
//   
//==========================================================================================

// 
extern struct VARS far * vp;

#pragma asm 
PUBLIC _VARD
#pragma endasm
volatile static const struct PVARD VARD={AllHeapsPointers+17 //MMI
  ,sizeof(struct VARS),"MAPVIEW0"};

_inline struct VARS far * GetVP(void)
{
 return(vp);
}
//==========================================================================================
//==========================================================================================
//==========================================================================================
//      cell id
//==========================================================================================

void loadCellID(void)
{
 unsigned int base;
 unsigned int ci;
 unsigned int ci0;
 unsigned int lac;
 unsigned int rci=0;
 unsigned int rlac=0;
 unsigned int x=0;
 unsigned int y=0;

 char s[256];
 int f;
 unsigned int i;
 unsigned int state=0;
 char c;
 unsigned int oldbase;

 struct VARS far *VP=GetVP();

 ci=VP->loadci;
 ci0=(ci/10)*10;
 lac=VP->loadlac;
 base=VP->bs_cp;
 VP->bs[base].x=0;
 VP->bs[base].y=0;
 oldbase=0xFFFF; // 

 f=FileOpen((STR)BSDAT_FILENAME,_O_RDONLY,0);
 if (f==-1) goto L_ZHOPA;
 //*(int *)&update_scr=2; // 
 L1:
 i=FileRead(f,s,sizeof(s));
 if (i<sizeof(s))
 {
  s[i]=0; //EOF
 }
 i=0;
 for(;;)
 {
  if (i==sizeof(s)) goto L1; // 
  c=s[i];
  if (!c) break; // 
  i++;
  if ((c==13)||(c==10))
  {
   state=0;
   rlac=0;
   rci=0;
   x=0;y=0;
   continue;
  }
  switch(state)
  {
   case 0:
          if ((c<'0')||(c>'9')) state++;
		  else rlac=rlac*10+(c-'0');
		  break;
   case 1:
          if ((c<'0')||(c>'9')) state++;
		  else rci=rci*10+(c-'0');
		  break;
   case 2:
          if ((c<'0')||(c>'9')) state++;
		  else x=x*10+(c-'0');
		  break;
   case 3:
          if ((c<'0')||(c>'9')) state++;
		  else y=y*10+(c-'0');
		  break;
   case 4:
          if (rlac==lac)
          {
		   if (rci==ci)
		   {
			// 
			goto L2;
           }
           if (rci==ci0)
           {
           	//      
			if (oldbase==0xFFFF) goto L2;
			//Ec  (, )
			if (x&&y)
			{
			 // 
             VP->bs[base].x=x;
             VP->bs[base].y=y;
			}
           }
          }
		  state=100;
		  break;
		  L2:
          VP->bs[base].ci=ci;
          VP->bs[base].lac=lac;
		  if (x&&y)
		  {
           VP->bs[base].x=x;
           VP->bs[base].y=y;
		  }
		  oldbase=base;
   default:
		  state++;
          if (state<36)
          { 
           VP->bs[base].s[state-5]=c;
           VP->bs[base].s[state-4]=0;
          }
		  break;
   case 100:
          break;
  }
 }
 FileClose(f);
 L_ZHOPA:
 if (oldbase==0xFFFF)
 {
  //  
  VP->bs[base].lac=lac;
  VP->bs[base].ci=ci;
  VP->bs[base].s[0]=0; // 
 }
 base++;
 if (base>=BS_CACHE_SIZE) base=0;
 VP->bs_cp=base;
 VP->loadCI_in_progress=0;
}

void WriteLogCI(void)
{
 int f;
 struct VARS far *VP=GetVP();
 f=FileOpen((STR)BSLOG_FILENAME,_O_CREAT+_O_RDWR+_O_APPEND,_S_IREAD);
 if (f==-1) return;
 FileWrite(f,VP->aCIupdate,strlen(VP->aCIupdate));
 FileClose(f);
 VP->loadCI_in_progress=0;
}

//==========================================================================================
//  cell  
//==========================================================================================
struct BS far * FindCell(unsigned int lac, unsigned int ci)
{
 struct VARS far *VP=GetVP();
 unsigned int base;
 unsigned int i;
 base=0xFFFF; //  
 //     
 i=0;
 do
 {
  if (lac!=VP->bs[i].lac) continue;
  if (VP->bs[i].ci==ci)
  {
   // /
   base=i;
   break;
  }
 }
 while(++i<BS_CACHE_SIZE);
 if (base==0xFFFF)
 {
  //   ,    
  VP->loadlac=lac;
  VP->loadci=ci;
  FilesysICall(loadCellID);
  return(0);
 }
 return((VP->bs)+base);
}

//==========================================================================================
//  cell id
//==========================================================================================
void WorkCellID(void)
{
 unsigned int ci;
 unsigned int lac;
 struct BS far * base;
 struct VARS far *VP=GetVP();

 if (VP->loadCI_in_progress) return;
 VP->loadCI_in_progress=1; //
 ci=NETMON_CI;
 lac=NETMON_LAC;

 if ((ci==VP->oldci)&&(lac==VP->oldlac)) goto L_NOCHANGE; //  

 base=FindCell(lac,ci);
 if (!base) return; //   ,  

 xsprintf(VP->aCIupdate,"%02d:%02d %02d-%02d-%04d    %u:%u->%u:%u (%s)->(%s)\r\n",
          _hour,_minute,_day,_month,_year,
          VP->oldlac,VP->oldci,lac,ci,VP->celltext,base->s
		  );
 FilesysICall(WriteLogCI);
 VP->oldci=ci;
 VP->oldlac=lac;
 if (base->s[0])
 {
  xsprintf(VP->celltext,"%s",base->s);
 }
 else
 {
  xsprintf(VP->celltext,"%u:%u???",lac,ci);
 }
 if (VP->update_scr==0) VP->update_scr=1;
 L_NOCHANGE:
 VP->loadCI_in_progress=0;
}

//==========================================================================================
//==========================================================================================
void reload_map(void)
{
 //Loading map to buffer
 //calling throws FilesysICall
 int f;
 unsigned int x,y;
 long fp;
 int i;
 struct VARS far *VP=GetVP();

 f=FileOpen((STR)MAP_FILENAME,_O_RDONLY,0);
 if (f==-1) return;
 FileRead(f,&VP->xsize,2);
 FileRead(f,&fp,2);
 FileRead(f,&VP->ysize,2);
 VP->ysize=VP->ysize&0xFFF8; //  8
 x=VP->mapx;
 y=VP->mapy+8; // 
 i=0;
 do
 {
  FileSeek(f,(unsigned long)x*VP->ysize+y,0,&fp);
  if (VP->mem8) FileRead(f,(VP->mem8)+(i<<8),256);
  x++;
  i++;
 }
 while(i<MAP_W);
 FileClose(f);
 VP->update_scr=1; // 
}

unsigned int DrawOneCell(struct CELLDATA far *cd)
{
 struct BS far * base;
 struct VARS far *VP=GetVP();
 unsigned int k;
 unsigned int lac=cd->LAC;
 unsigned int ci=cd->CI;
 unsigned int rxlvl=GetRXLVL_by_CH(cd->CH);
 lac=(lac<<8)|(lac>>8);
 if (!ci) return(0); // 
 if (!rxlvl) return(0); // 
 k=0;
 while(k<VP->c3draw)
 {
  base=VP->BS7[k];
  if ((base->lac==lac)&&(base->ci==ci)) return(0); // 
  k++;
 }
 base=FindCell(lac,ci); if (!base) return(1); //  
 if (base->x&&base->y)
 {
  //   
  k=0;
  if (rxlvl<(sizeof(PWRS)/sizeof(unsigned int))) k=PWRS[rxlvl];
  VP->sumk+=k;
  VP->sumx+=_mulu32(base->x,k);
  VP->sumy+=_mulu32(base->y,k);
 }
 base->lvl=rxlvl;
 //xsprintf(VP->celltext2+(VP->c3draw*7),"%04u.%01u\n",ci/10,ci%10);
 VP->BS7[VP->c3draw]=base;
 VP->c3draw++;
 return(0);
}

int Timer_drawmap(void)
{
 char rm;
 unsigned int x;
 unsigned int y;
 char far *d;
 char far *s;
 int i;
 struct rectXYXY rc;
 struct VARS far *VP=GetVP();
 unsigned int gx;
 unsigned int gy;

 rc.X1=0;
 rc.X2=100;

 if ((VP->manual_x)||(VP->manual_y))
 {
  gx=VP->manual_x>>3;
  gy=VP->manual_y>>3;
  VP->manual_x=VP->manual_y=0;
  goto L_DIRECT_LOAD_MAP;
 }
 if (VP->update_scr!=1) return(1); //   
 rm=VP->rollmode;
 if ((rm=='5')||(rm=='4'))
 {
  char s[50];
  //  
  if (VP->loadCI_in_progress) return(1); // 
  VP->loadCI_in_progress=1; //
  i=VP->cdraw;
  switch(i)
  {
   case 7:
    //  
	x=VP->sumk;
	if (x)
	{
	 VP->gpsx=_divu32(VP->sumx,x);
	 VP->gpsy=_divu32(VP->sumy,x);
    }
    if (rm!='5') break;
    FillRect(0,0,101,80,0);
	i=0;
	do
	{
	 if (!(VP->BS7[i])) break;
     xsprintf(s,"%04u.%01u:%s",VP->BS7[i]->ci/10,VP->BS7[i]->ci%10,VP->BS7[i]->s);
	 rc.Y2=(rc.Y1=i*11)+8;
     VP->up=VP->us;
     STRtoWSTRP(&VP->up,s);
     PrepDrawObj_type01(VP->dobj,&rc,0,&VP->up,0,0x80);
     DrawObject(VP->dobj);
	 x=105-VP->BS7[i]->lvl;
     if (x&0x8000) x=0;
     x<<=1;
     if (x>100) x=100;
	 FillRect(0,rc.Y1+8,x,2,1);
     i++;
    }
	while(i<7);
	break;
   case 10:
   case 0:
    VP->cdraw=0;
    VP->c3draw=0;
	VP->sumx=0;
	VP->sumy=0;
	VP->sumk=0;
	memset(VP->BS7,0,sizeof(VP->BS7));
    if (DrawOneCell(CURCELL_PNT)) return(1);
    break;
   case 1:
   case 2:
   case 3:
   case 4:
   case 5:
   case 6:
    if (DrawOneCell(CELL6_PNT[i-1])) return(1);
	break;
  }
  VP->cdraw++;
  VP->loadCI_in_progress=0; // 
  if (rm=='5') return(1);
 }
 if (rm=='4')
 {
  gx=VP->gpsx>>3;
  gy=VP->gpsy>>3;
  rm=0;
  //xsprintf(VP->aCRD,"Track %05u %05u",VP->gpsx,VP->gpsy);
  FillRect(0,70,101,10,VP->track_upd?0:2);
  //DrawString(0,71,101,9,VP->aCRD,0); 
  VP->up=VP->us;
  STRtoWSTRP(&VP->up,VP->celltext);
  rc.Y1=71;
  rc.Y2=100;
  PrepDrawObj_type01(VP->dobj,&rc,0,&VP->up,0,0x81);
  DrawObject(VP->dobj);

  x=VP->viewx+VP->mapx+(SCR_W/2); //
  y=((VP->viewy+VP->mapy)>>3)+(SCR_H/16); //
  //,     
  i=x-gx;
  if (i>MAX_DELTA||i<-MAX_DELTA) goto L_DIRECT_LOAD_MAP;
  i=y-gy;
  if (i>MAX_DELTA||i<-MAX_DELTA)
  {
   L_DIRECT_LOAD_MAP:
   gy<<=3;
   if (gx>=SCR_W/2) gx-=SCR_W/2; else gx=0; //X  
   if (gy>=SCR_H/2) gy-=SCR_H/2; else gy=0; //Y  
   if (gx>(VP->xsize-SCR_W)) gx=VP->xsize-SCR_W; // x  
   if (gy>(VP->ysize-SCR_H)) gy=VP->ysize-SCR_H; // y  
   // gx  gy -     
   if (gx<(VP->xsize-MAP_W))
   {
    if (gx>=(MAP_W/2))
     VP->mapx=gx-(VP->viewx=MAP_W/2);
    else
     {VP->mapx=0; VP->viewx=gx;}
   }
   else VP->viewx=gx-(VP->mapx=VP->xsize-MAP_W);

   if (gy<(VP->ysize-SCR_H))
   {
    if (gy>=(MAP_H/2))
     VP->mapy=gy-(VP->viewy=MAP_H/2);
    else
     {VP->mapy=0; VP->viewy=gy;}
   }
   else VP->viewy=gy-(VP->mapy=VP->ysize-MAP_H);
   VP->track_upd=0;
   VP->update_scr=2;
   FilesysICall(reload_map); // 
   return(1);
  }
  if (VP->cdraw&1)
  {
   if (gx>x) rm=0x3E; // 
   if (gx<x) rm=0x3D; //
  }
  else
  {
   if (gy>y) rm=0x3C; //
   if (gy<y) rm=0x3B; //
  }
  if (rm) VP->track_upd=0; // 
  if (VP->track_upd) return(1); // , 
 }
 else 
 {
  FillRect(0,70,101,10,0);
  VP->up=VP->us;
  STRtoWSTRP(&VP->up,VP->celltext);
  rc.Y1=71;
  rc.Y2=100;
  PrepDrawObj_type01(VP->dobj,&rc,0,&VP->up,0,0x81);
  DrawObject(VP->dobj);
  VP->update_scr=0; // 
 }
 FillRect(96,0,5,64,0); //   
 FillRect(0,64,101,6,0); //   
 x=_divu32(_mulu32(88,VP->viewx+VP->mapx),VP->xsize);
 y=_divu32(_mulu32(56,VP->viewy+VP->mapy),VP->ysize);
 FillRect(x,65,8,2,1);
 FillRect(97,y,2,8,1);
 x=VP->viewx;
 y=VP->viewy;
 //d=VP->scr;
 if (VP->mem8)
 {
  d=(char far *)(0x0E*0x4000L+0x160C); //0E:160C -  
  x<<=8;
  x+=y;
  s=VP->mem8+x;
  i=SCR_H;
  do
  {
   d[0]=s[0x0000];
   d[1]=s[0x0100];
   d[2]=s[0x0200];
   d[3]=s[0x0300];
   d[4]=s[0x0400];
   d[5]=s[0x0500];
   d[6]=s[0x0600];
   d[7]=s[0x0700];
   d[8]=s[0x0800];
   d[9]=s[0x0900];
   d[10]=s[0x0A00];
   d[11]=s[0x0B00];
   d+=13;
   s++;
  }
  while(--i);
  // 
  d=(char far *)(0x0E*0x4000L+0x160C+5+(24*13)); //0E:160C -  
  i=16;
  s=(char far *)((VP->rollmode=='4')?sprites:sprites+64);
  do
  {
   char c;
   c=*d;
   c&=*s++;
   c|=*s++;
   *d=c;
   d++;
   c=*d;
   c&=*s++;
   c|=*s++;
   *d=c;
   d+=12;
  }
  while(--i);
  InvalidateScreenRect(0,0,SCR_W*8,SCR_H);
 }
 else
 {
  DrawString(0,0,96,64,(STR)"Out of memory for map buffer!",3);
 }
 VP->track_upd=1;

 x=VP->viewx;
 switch(rm)
 {
  //Left
  case LEFT_BUTTON:
   if (x==0) break; //   
   x--;
   if (x==0)
   {
    // ,   
	x=VP->mapx;
	if (x>0) //     
	{
	 i=x;
	 if (x<MAP_W/2) x=0; else x-=MAP_W/2;
	 i=i-x; //    
	 VP->mapx=x; //  x   
	 x=i; //  x  
	 VP->update_scr=2;
     FilesysICall(reload_map); // 
	}
   }
   break;
  //Right
  case RIGHT_BUTTON:	
   if (x==(MAP_W-SCR_W)) break; //  
   x++;
   if (x==(MAP_W-SCR_W))
   {
    // ,   
	x=VP->mapx;
	if (x<(VP->xsize-MAP_W)) //     
	{
	 i=x;
	 x+=MAP_W/2;
	 if (x>(VP->xsize-MAP_W)) x=VP->xsize-MAP_W;
	 i=x-i; //    
	 VP->mapx=x; //  x   
	 x=(MAP_W-SCR_W)-i; //  x  
	 VP->update_scr=2;
     FilesysICall(reload_map); // 
	}else x=(MAP_W-SCR_W); //
   }
   break;
  //Up
  case UP_BUTTON:	
   if (y==0) break; //   
   y-=8;
   if (y==0)
   {
    // ,   
	y=VP->mapy;
	if (y>0) //     
	{
	 i=y;
	 if (y<MAP_H/2) y=0; else y-=MAP_H/2;
	 i=i-y; //    
	 VP->mapy=y; //  y   
	 y=i; //  y  
	 VP->update_scr=2;
     FilesysICall(reload_map); // 
	}
   }
   break;
  //Down
  case DOWN_BUTTON:
   if (y==(MAP_H-SCR_H)) break; // 
   y+=8;
   if (y==(MAP_H-SCR_H))
   {
    // ,   
	y=VP->mapy;
	if (y<(VP->ysize-MAP_H)) //     
	{
	 i=y;
	 y+=MAP_H/2;
	 if (y>(VP->ysize-MAP_H)) y=VP->ysize-MAP_H;
	 i=y-i; //    
	 VP->mapy=y; //  x   
	 y=(MAP_H-SCR_H)-i; //  x  
	 VP->update_scr=2;
     FilesysICall(reload_map); // 
	}else y=(MAP_H-SCR_H); //
   }
   break;
   default:
   // 
   return(1);
 }
 VP->viewx=x;
 VP->viewy=y;
 VP->track_upd=0;
 if (VP->update_scr==0) VP->update_scr=1; //   . 
 return(1);
}

void diaCreator(void)
{
 char db[0x2C]; 
 //doIDLE();
 CreateDialogWithSoftKeys(&maindialog,db,0);
}

void IdleHook(void)
{
 struct VARS far *VP=GetVP();
 int i=VP->old_lock_mode;
 unsigned int m=0;
 WorkCellID();
 if (i==-1) goto L_NORUN;
 if ((P3&0x0280)==0)
 {
  m=1;
  goto L_2KEY;
 }
 if (!GetKeyLockState1())
 {
  L_2KEY:
  if (i)
  {
   i--;
  }
  else
  {
   VP->old_lock_mode=i=-1; //   
   VP->update_scr=-1; // 
   VP->track_upd=0;
   FilesysICall(diaCreator);
  }
 }
 else i=0;
 VP->old_lock_mode=i;
 L_NORUN:
 DPP0=0;
}

void DO_AddBS(void)
{
 int f;
 char s[100];
 struct VARS far *VP=GetVP();

 f=FileOpen((STR)BSDAT_FILENAME,_O_CREAT+_O_RDWR+_O_APPEND,_S_IREAD);
 if (f==-1) return;
 xsprintf(s,"%u\t%u\t%u\t%u\tAdded by MAP/GPS\r\n",NETMON_LAC,(NETMON_CI/10)*10,
  (VP->viewx+VP->mapx+6)*8,
  (VP->viewy+VP->mapy+32)
  );
 FileWrite(f,s,strlen(s));
 FileClose(f);
}

onKey_AddBS(void far *data, int keycode)
{
 switch(keycode)
 {
  case 0x23:
   FilesysICall(DO_AddBS);
  case 0x24:
   //LockKBD();
   break;
 }
 return(keycode);
}

void AskAboutAddBS(void)
{
 static const int txt[]={0x1D1,0x7FFF};
 static const struct SKDEFINE btn[]=
 {
  {0x23,0x52,0x152},
  {0x52,0x52,0xFFFE},
  {0x52,0x52,0xFFFE},
  {0x24,0x52,0x1C0}
 };
 static const struct LOCMESS_I lm=
 {
  3,0,2,
  30000,
  0,
  0x7FFF,
  0,
  onKey_AddBS,
  0,
  (struct SKDEFINE far *)&btn,
  20006,
  2,
  2
 };
 ShowLocalizedMessageWithIconAndNoBorder(&lm, txt);
}

#define mf_ENTRY(VAL) (VP->mem8[36*(VAL)])

void manual_load(int dia_id)
{

 char s[256];
 int f;
 unsigned int i;
 unsigned int state;
 char c;
 unsigned int lp;
 unsigned int dp; //    
 char fs[32]; // 
 char ks[32]; //  
 unsigned int sp;

 unsigned int x;
 unsigned int y;

 struct VARS far *VP=GetVP();

 L_RELOAD:
 x=0;
 y=0;
 state=0;
 VP->mf_state=1;
 memset(VP->mem8,0,8192);
 VP->mf_pos=0;
 SendMessage((void far *)0x3B5E8,0xF8,dia_id); //
 //     
 f=FileOpen((STR)BSDAT_FILENAME,_O_RDONLY,0);
 lp=0;
 if (f==-1) goto L_ZHOPA;
 L1:
 i=FileRead(f,s,sizeof(s));
 if (i<sizeof(s))
 {
  s[i]=0; //EOF
 }
 i=0;
 for(;;)
 {
  if (VP->mf_state==2)
  {
   FileClose(f);
   goto L_RELOAD;
  }
  if (i==sizeof(s)) goto L1; // 
  c=s[i];
  if (!c) break; // 
  i++;
  if ((c==13)||(c==10))
  {
   if (state>4)
   {
	// 
    AcquireGbsLock();
	//  
	sp=0;
	dp=0;
	while((c=VP->digits[dp]))
	{
	 if (c==ks[sp+dp]) dp++; else
	 {
	  dp=0;
	  sp++;
	  if (ks[sp]==0) break; 
	 }
	}
	if (VP->mf_state==2) c=1; //   
	if (!c)
	{
	 memcpy(VP->mem8+lp,fs,state-4);
	 ((unsigned int *)(VP->mem8+lp))[16]=x;
	 ((unsigned int *)(VP->mem8+lp))[17]=y;
	 lp+=36; // 
	}
    FreeGbsLock();
    if (!c) SendMessage((void far *)0x3B5E8,0xF8,dia_id); //
	if (lp>=8000) break;
   }
   state=0;
   x=0;y=0;
   continue;
  }
  switch(state)
  {
   case 0:
   case 1:
          if ((c<'0')||(c>'9')) state++;
		  break;
   case 2:
          if ((c<'0')||(c>'9')) state++;
		  else x=x*10+(c-'0');
		  break;
   case 3:
          if ((c<'0')||(c>'9')) state++;
		  else y=y*10+(c-'0');
		  break;
   case 4:
          //   
          memset(ks,0,sizeof(ks)); // 
   default:
          sp=state-4;
          if (sp<31)
		  {
		   state++;
		   fs[sp]=c; //
		   ks[sp]=mf_table[c]; //
          }
		  break;
   case 100:
          break;
  }
 }
 FileClose(f);
 if (VP->mf_state==2) goto L_RELOAD;
 L_ZHOPA:
 VP->mf_state=0;
 SendMessage((void far *)0x3B5E8,0xF8,dia_id);
}

void mf_redraw(void)
{
 unsigned int i;
 struct VARS far *VP=GetVP();
 struct rectXYXY rc;
 rc.X1=0;
 rc.X2=100;
 rc.Y2=(rc.Y1=12)+12;
 DrawString(0,0,101,12,"Search key:",3); 
 VP->up=VP->us;
 STRtoWSTRP(&VP->up,(STR)VP->digits);
 PrepDrawObj_type01(VP->dobj,&rc,0x40,&VP->up,3,0x80);
 DrawObject(VP->dobj);
 if (VP->mf_pos>=3) i=VP->mf_pos-3; else i=0;
 rc.Y2=(rc.Y1=80-(7*8))+8;
 while(rc.Y2<=80)
 {
  if (!mf_ENTRY(i))
  {
   FillRect(0,rc.Y1,101,8,0);
  }
  else
  {
   STRtoWSTRP(&VP->up,&mf_ENTRY(i));
   PrepDrawObj_type01(VP->dobj,&rc,i==VP->mf_pos?0x60:0x40,&VP->up,0,0x80);
   DrawObject(VP->dobj);
  }
  rc.Y1+=8;
  rc.Y2+=8;
  i++;
 }
}

int mf_onKey(void far *data, struct MSG far *Msg)
{
 struct VARS far *VP=GetVP();

 if ((Msg->Msg==0xF8)&&(Msg->Param[0]==((int far *)data)[6]))
 {
  if (IsFocus(data)) mf_redraw();
  return(0);
 }
 if (!IsFocus(data)) return(1);
 if ((Msg->Msg==0x90)||(Msg->Msg==0x88)) //  . 
 {
  if (VP->mf_pos!=-1) mf_redraw();
  return(0);
 }
 if (Msg->Msg==KEY_UP) return(0);
 if ((Msg->Msg==KEY_DOWN)||(Msg->Msg==LONG_PRESS))
 {
  switch(Msg->Param[0])
  {
   case GREEN_BUTTON:
    VP->manual_x=((unsigned int *)(VP->mem8+(VP->mf_pos*36)))[16];
    VP->manual_y=((unsigned int *)(VP->mem8+(VP->mf_pos*36)))[17];
    if (VP->manual_x||VP->manual_y)
	{
	 VP->rollmode=0;
    }
	else break;
   case RED_BUTTON:
    VP->update_scr=2;
    FilesysICall(reload_map); // 
    CloseDialogWithSoftKeys(((int far *)data)[6]);
	break;
   case UP_BUTTON:
    if (VP->mf_pos==-1) break;
    if (VP->mf_pos)
	{
	 VP->mf_pos--;
	 mf_redraw();
    }
	break;
   case DOWN_BUTTON:
    if (VP->mf_pos==-1) break;
    if (mf_ENTRY(VP->mf_pos)&&mf_ENTRY(VP->mf_pos+1))
    {
     VP->mf_pos++;
	 mf_redraw();
	}
	break;
   case LEFT_SOFT:
    if (strlen((STR)(VP->digits))) VP->digits[strlen((STR)(VP->digits))-1]=0;
	goto L_RELOAD;
   case '0': case '1': case '2': case '3': case '4': case '5':
   case '6': case '7': case '8': case '9': case '*': case '#': 
    //   
    if (strlen((STR)(VP->digits))<(sizeof(VP->digits)-1)) strcat((STR)(VP->digits),(STR)(Msg->Param));
	L_RELOAD:
	if (VP->mf_state==0)
	{
	 VP->mf_state=1;
	 FilesysICall_I(manual_load,((int far *)data)[6]);
	}
	else
	{
	 VP->mf_state=2; //
    }
	break;
   default:
    break;
  }
  return(0);
 }
 return(1);
}

void mf_onCreate(void far *data)
{
 struct VARS far *VP=GetVP();
 VP->mf_pos=-1;
 VP->mf_state=1;
 FillRect(0,0,101,80,0);
 memset(VP->digits,0,sizeof(VP->digits));
 ((unsigned int far *)data)[7]=0;
 FilesysICall_I(manual_load,((int far *)data)[6]); // 
}

void mf_onClose(void)
{
}

void ManualFind(void)
{
 static const struct DIALOG_WSK manualfinddialog=
 {
  mf_onKey,
  mf_onCreate,
  mf_onClose,
  0x2C,
  sizeof(md_tab)/sizeof(unsigned int),
  md_tab
 };
 char db[0x2C]; 
 CreateDialogWithSoftKeys(&manualfinddialog,db,0);
}

int md_onKey(void far *data, struct MSG far *Msg)
{
 struct VARS far *VP=GetVP();
 //char c;
 if (!IsFocus(data))
 {
  if (VP->tmr_online)
  {
   DisableTimerX(&VP->tmr);
   VP->tmr_online=0;
  } 
  return(1);
 }
 //if (Msg->Msg==0xB7) return(0); //  
 if ((Msg->Msg==0x90)||(Msg->Msg==0x88)) //  . 
 {
   //char s[32];
   StartTimer(&VP->tmr,26,0x92);
   VP->tmr_online=1;
   if (VP->update_scr==0) VP->update_scr=1;
   VP->track_upd=0;
   Timer_drawmap();
   //xsprintf(s,"MSG: %02X",Msg->Msg);
   //DrawString(0,72,101,8,s,0);
   return(0);
 }
 if (Msg->Msg==0x92)
 {
  VP->tmr_online=1;
  StartTimer(&VP->tmr,26,0x92);
  Timer_drawmap();
  return(0);
 }
 if (Msg->Msg==KEY_DOWN)
 {
  turnLightON();
  switch(Msg->Param[0])
  {
   case '1':
   case '2':
   case '3':
    SetBAND(Msg->Param[0]-0x30);
	break;
   case '4': 
    if (VP->rollmode=='4') VP->rollmode=0; else 
    {
     VP->rollmode='4';
	 VP->track_upd=0;
	}
	break;
   case '5': 
    if (VP->rollmode=='5') VP->rollmode=0; else VP->rollmode='5';
	break;
   case '6':
    DisableTimerX(&VP->tmr);
    VP->tmr_online=0;
    //CloseDialogWithSoftKeys(((int far *)data)[6]);
    AskAboutAddBS();
	return(0);
   case '0':
    if (VP->mem8) //  
	{
     DisableTimerX(&VP->tmr);
	 VP->tmr_online=0;
	 ManualFind();
	}
	return(0);
   default:
    if (VP->rollmode) break;
    VP->rollmode=Msg->Param[0]; // 
	break;
  }
  if (VP->update_scr==0) VP->update_scr=1;
  return(0);
 }
 if (Msg->Msg==KEY_UP)
 {
  if (VP->ready_for_exit&&(Msg->Param[0]=='*'))
  {
   DisableTimerX(&VP->tmr);
   CloseDialogWithSoftKeys(((int far *)data)[6]);
   return(0);
  }
  if ((VP->rollmode!='4')&&(VP->rollmode!='5')) VP->rollmode=0;
  if (VP->update_scr==0) VP->update_scr=1;
  return(0);
 }
 if (Msg->Msg==LONG_PRESS)
 {
  if (Msg->Param[0]=='*')
  {
   VP->ready_for_exit=1;
  }
  else VP->ready_for_exit=0;
  return(0);
 }
 return(1);
}


void sd_oncreate(void)
{
 doBack0A();
}

void sd_onkey(void)
{
}

void sd_oninit(void)
{
}

void sd_onexit1(void)
{
}

void sd_onexit2(void)
{
}

static const struct DIALOG sdialog={sd_onkey,sd_oninit,sd_onexit1,sd_onexit2,sd_oncreate,0x10,0xc4};

void md_onCreate(void far *data)
{
 char mmm[0x10];
 struct VARS far *VP=GetVP();

 VP->ready_for_exit=0;

 UnLockKBD();
 pShowDialog(&sdialog,mmm);

 FillRect(0,0,101,80,0);
 //VP->oldci=0;
 //VP->oldlac=0;
 //FilesysICall(save_dump);
 DrawString(0,0,101,80,"MAP/GPS v1.4\n(C)2005/2006 by Rst7/CBSIE\nUse cursor to move.",3);
 VP->update_scr=2;
 VP->mem8=EX_heap_malloc_with_lock(8192);
 FilesysICall(reload_map);
 OnTimer1();
 OnTimer2();
 StartTimer(&VP->tmr,26,0x92);
 ((unsigned int far *)data)[7]=0;
}

/*void dummy_timer(void)
{
 struct VARS far *VP=GetVP();
 //SendMessage((void far *)0x3B5E8,0x90,0);
 //SendMessage((void far *)0x3B5E8,0x88,0);
 DisplayMessageBoxV(1,0x400);
 DisableTimerX(&VP->tmr);
}*/

void md_onClose(void)
{
 struct VARS far *VP=GetVP();
 DisableTimerX(&VP->tmr);
 //StartTimer2(&VP->tmr,20,dummy_timer);
 if (VP->mem8) EX_heap_free_with_lock(VP->mem8);
 VP->old_lock_mode=1000;
}
